use std::hash::{DefaultHasher, Hash, Hasher};

// TODO: Documentation pass!

/* -------------------------------------------------------------------------- */
/*                                  Interface                                 */
/* -------------------------------------------------------------------------- */

pub trait AuthKey: PartialEq {}

/// AuthGuards serve as a way to lock certain resources behind a requirement
pub trait AuthGuard {
    type Key: AuthKey;
    fn verify(&self, key: &Self::Key) -> Result<bool, ()>;
}

/* ------------------------------- Composition ------------------------------ */
pub struct ComposeAndAuthGuard<A, B>
where
    A: AuthGuard,
    B: AuthGuard<Key = A::Key>,
{
    pub left: A,
    pub right: B,
}

impl<A, B> AuthGuard for ComposeAndAuthGuard<A, B>
where
    A: AuthGuard,
    B: AuthGuard<Key = A::Key>,
{
    type Key = A::Key;

    fn verify(&self, key: &Self::Key) -> Result<bool, ()> {
        // Return false as soon as one of the guards fails
        if !self.left.verify(key)? {
            return Ok(false);
        }
        self.right.verify(key)
    }
}

pub struct ComposeOrAuthGuard<A, B>
where
    A: AuthGuard,
    B: AuthGuard<Key = A::Key>,
{
    pub left: A,
    pub right: B,
}

impl<A, B> AuthGuard for ComposeOrAuthGuard<A, B>
where
    A: AuthGuard,
    B: AuthGuard<Key = A::Key>,
{
    type Key = A::Key;

    fn verify(&self, key: &Self::Key) -> Result<bool, ()> {
        // Return true as soon as one of the guards succeeds
        if self.left.verify(key)? {
            return Ok(true);
        }
        self.right.verify(key)
    }
}

/* ------------------------------ Unit Tests ------------------------------ */
#[cfg(test)]
mod tests {
    use super::*;

    // A dummy key that implements AuthKey.
    #[derive(Debug, PartialEq)]
    struct DummyKey(u64);

    impl AuthKey for DummyKey {}

    // A mock AuthGuard that always returns true.
    struct AlwaysTrue;

    // A mock AuthGuard that always returns false.
    struct AlwaysFalse;

    impl AuthGuard for AlwaysTrue {
        type Key = DummyKey;

        fn verify(&self, _key: &Self::Key) -> Result<bool, ()> {
            Ok(true)
        }
    }

    impl AuthGuard for AlwaysFalse {
        type Key = DummyKey;

        fn verify(&self, _key: &Self::Key) -> Result<bool, ()> {
            Ok(false)
        }
    }

    #[test]
    fn test_and_auth_guard_both_true() {
        let guard = ComposeAndAuthGuard {
            left: AlwaysTrue,
            right: AlwaysTrue,
        };
        let key = DummyKey(42);
        assert_eq!(guard.verify(&key).unwrap(), true);
    }

    #[test]
    fn test_and_auth_guard_true_false() {
        let guard = ComposeAndAuthGuard {
            left: AlwaysTrue,
            right: AlwaysFalse,
        };
        let key = DummyKey(42);
        assert_eq!(guard.verify(&key).unwrap(), false);
    }

    #[test]
    fn test_and_auth_guard_false_true() {
        let guard = ComposeAndAuthGuard {
            left: AlwaysFalse,
            right: AlwaysTrue,
        };
        let key = DummyKey(42);
        assert_eq!(guard.verify(&key).unwrap(), false);
    }

    #[test]
    fn test_and_auth_guard_both_false() {
        let guard = ComposeAndAuthGuard {
            left: AlwaysFalse,
            right: AlwaysFalse,
        };
        let key = DummyKey(42);
        assert_eq!(guard.verify(&key).unwrap(), false);
    }

    #[test]
    fn test_or_auth_guard_both_true() {
        let guard = ComposeOrAuthGuard {
            left: AlwaysTrue,
            right: AlwaysTrue,
        };
        let key = DummyKey(42);
        assert_eq!(guard.verify(&key).unwrap(), true);
    }

    #[test]
    fn test_or_auth_guard_true_false() {
        let guard = ComposeOrAuthGuard {
            left: AlwaysTrue,
            right: AlwaysFalse,
        };
        let key = DummyKey(42);
        assert_eq!(guard.verify(&key).unwrap(), true);
    }

    #[test]
    fn test_or_auth_guard_false_true() {
        let guard = ComposeOrAuthGuard {
            left: AlwaysFalse,
            right: AlwaysTrue,
        };
        let key = DummyKey(42);
        assert_eq!(guard.verify(&key).unwrap(), true);
    }

    #[test]
    fn test_or_auth_guard_both_false() {
        let guard = ComposeOrAuthGuard {
            left: AlwaysFalse,
            right: AlwaysFalse,
        };
        let key = DummyKey(42);
        assert_eq!(guard.verify(&key).unwrap(), false);
    }
}

/* -------------------------------------------------------------------------- */
/*                           Default Implementations                          */
/* -------------------------------------------------------------------------- */

#[derive(Debug, Clone)]
pub struct CodeAuthKey {
    key: u64
}

impl CodeAuthKey {
    fn _generate_key<T: Hash + ?Sized>(code: &T) -> u64 {
        let mut s = DefaultHasher::new();
        code.hash(&mut s);
        s.finish()
    }

    pub fn new<T: Hash + ?Sized>(code: &T) -> CodeAuthKey {
        CodeAuthKey {
            key: Self::_generate_key(code)
        }
    }
}

impl AuthKey for CodeAuthKey {}

impl PartialEq for CodeAuthKey {
    fn eq(&self, other: &Self) -> bool {
        return self.key == other.key;
    }
}

pub struct CodeAuthGuard {
    reference_key: CodeAuthKey
}

impl CodeAuthGuard {
    fn new(key: CodeAuthKey) -> Self {
        Self {
            reference_key: key
        }
    }
}

impl AuthGuard for CodeAuthGuard {
    type Key = CodeAuthKey;
    
    fn verify(&self, key: &Self::Key) -> Result<bool, ()> {
        Ok(self.reference_key == *key)
    }
}

/* ------------------------------- Unit Tests ------------------------------- */
#[cfg(test)]
mod code_authguard_tests {
    use rand::{distr::Alphanumeric, prelude::*};
    use stopwatch::Stopwatch;
    use super::*;

    fn _create_rng_string() -> String {
        let mut rng = rand::rng();
        let count = rng.random_range(5..100);
        rng
            .sample_iter(Alphanumeric)
            .take(count)
            .map(char::from)
            .collect()
    }

    #[test]
    fn basic_use() {
        let code_a = "reference_a";
        let code_b = "reference_b";
        
        // Using the keys directly
        let key_a = CodeAuthKey::new(&code_a);
        let key_b = CodeAuthKey::new(&code_b);

        assert_eq!(key_a, key_a);
        assert_eq!(key_b, key_b);
        assert_ne!(key_a, key_b);
        
        // Using the CodeAuthGuard
        let guard = CodeAuthGuard::new(key_a.clone());
        let mut result_verify = guard.verify(&key_a);
        assert!(result_verify.is_ok());
        assert!(result_verify.unwrap() == true);

        result_verify = guard.verify(&key_b);
        assert!(result_verify.is_ok());
        assert!(result_verify.unwrap() == false);
    }

    #[test]
    fn basic_benchmark_compare_speed() {
        // TODO: Look into better ways of performing benchmarks.
        
        let key_global = CodeAuthKey::new("reference");
        
        let count = 100000;

        let mut sw = Stopwatch::new();

        // Correct key test
        sw.start();

        for i in 0..count {
            let key_local = CodeAuthKey::new(&i);
            assert_eq!(key_local, key_local);
        }

        sw.stop();

        println!("Correct key compare perf: {}ms", sw.elapsed_ms());

        // Incorrect key test
        sw.reset();
        sw.start();

        for i in 0..count {
            let key_local = CodeAuthKey::new(&i);
            assert_ne!(key_local, key_global);
        }
        
        sw.stop();

        println!("Incorrect key compare perf: {}ms", sw.elapsed_ms());
    }
}